#!/usr/bin/python

import os
import time
import StringIO
from ConfigParser import RawConfigParser, MissingSectionHeaderError, NoSectionError
import subprocess
from socket import gethostname

from fjd.core_process import CoreProcess
from fjd.utils import ensure_wdir


class Worker(CoreProcess):
    '''
    A worker process
    '''

    def __init__(self, interval=.1, project=None):
        if not project:
            project = 'default'
        self.wdir = ensure_wdir(project)
        self.start_up()

        # announce my presence
        self.id = self.mk_id()
        print('[fjd-worker] Started with ID {id}.'.format(id=self.id))
        subprocess.call('touch {wdir}/workerqueue/{id}.worker'\
                   .format(wdir=self.wdir, id=self.id), shell=True)

        # look for jobs
        while True:
            job = self.next_job_on_pod()
            if job:
                print('[fjd-worker] Worker {}: I found a job.'.format(self.id))
                # Check if job file is a config file (ini-style).
                # If it is, get executable from there, call it and pass it the config file.
                # If it is not, run the job file as a script.

                # We read the file first and close it, so no stale handles will exists
                # in case ConfigParser exits
                with open('{}/jobpod/{}'.format(self.wdir, job), 'r') as jobfile:
                    jobtxt = jobfile.read()
                conf = RawConfigParser()
                ini_fp = StringIO.StringIO(jobtxt)
                try:
                    conf.readfp(ini_fp)  # this raises in case it is not an .ini file
                    exe = conf.get('fjd', 'executable')
                    cmd = 'nice -n {nice} {exe} {wdir}/jobpod/{job}; '\
                     .format(nice=9, exe=exe, wdir=self.wdir, job=job)
                except (MissingSectionHeaderError, NoSectionError):
                    cmd = 'nice -n {nice} {wdir}/jobpod/{job}'.format(nice=9, wdir=self.wdir, job=job)
                subprocess.call(cmd, shell=True)
                print('[fjd-worker] Worker {}: Finished my job.'.format(self.id))
                # remove the job from pod (signaling it is done) + re-announce myself
                subprocess.call('rm {wdir}/jobpod/{job}; touch {wdir}/workerqueue/{id}.worker'\
                        .format(wdir=self.wdir, job=job, id=self.id), shell=True)
            time.sleep(interval)

    def mk_id(self):
        ms = time.time()
        return '{h}_{t}'.format(h=gethostname(), t=ms)

    def next_job_on_pod(self):
        my_jobs = [j for j in os.listdir('{}/jobpod'.format(self.wdir))\
                  if j.startswith(self.id)]
        if len(my_jobs) > 0:
            return my_jobs[0]
        else:
            return None




import os
import os.path as osp


def ensure_wdir(project='default'):
    '''
    Makes sure working directory with needed sub-directories exists.
    Returns path to project directory.
    '''
    if project == '' or not project:
        project = 'default'
    fjd_home = osp.expanduser('~/.fjd')
    if not osp.exists(fjd_home):
        os.mkdir(fjd_home)
    if not osp.exists("{}/{}".format(fjd_home, project)):
        os.mkdir("{}/{}".format(fjd_home, project))
    for d in ('jobqueue', 'jobpod', 'workerqueue', 'screenrcs', 'screenlogs'):
        full_d = '{}/{}/{}'.format(fjd_home, project, d) 
        if not osp.exists(full_d):
            os.mkdir(full_d)
    return "{}/{}".format(fjd_home, project)


def empty_queues(project='default'):
    '''
    Empty the queues - creates a blank state
    '''
    if project == '' or not project:
        project = 'default'
    fjd_home = osp.expanduser('~/.fjd')
    if osp.exists(fjd_home):
        if osp.exists("{}/{}".format(fjd_home, project)):
            for d in ('jobqueue', 'jobpod', 'workerqueue', 'screenrcs', 'screenlogs'):
                full_d = '{}/{}/{}'.format(fjd_home, project, d) 
                for f in os.listdir(full_d):
                    os.remove("{}/{}".format(full_d, f))




#!/usr/bin/env python

import sys
import os
import time
import signal
import subprocess

from fjd import CoreProcess
from fjd import Recruiter
from fjd.utils import ensure_wdir


class Dispatcher(CoreProcess):
    '''
    Directory and File-based work dispatcher.
    Expects jobs to be put in directory `jobqueue` and workers to
    annouce themselves in directory `workerqueue`. Will then assign jobs 
    to workers by moving jobs to the `jobpod` directory. 
    '''

    def __init__(self, interval=.1, project=None, end_when_jobs_are_done=True,
                 end_callback=None, status_only=False):
        if not project:
            project = 'default'
        self.wdir = ensure_wdir(project)
        self.start_up()

        if not status_only:
            print('[fjd-dispatcher] Started on project "{}".'.format(project))

        def signal_handler(signal, frame):
            ''' gently exiting, e.g. when CTRL-C was pressed.  '''
            sys.stdout.write('\n[fjd-dispatcher] Received Exit signal. Exiting ...\n')
            print('[fjd-dispatcher] Should I fire all workers in project {}? [y|N]'\
                        .format(project))
            if raw_input().lower() in ["y", "yes"]:
                Recruiter(project=project).fire()
            sys.exit(0)
        signal.signal(signal.SIGINT, signal_handler)

        do_work = True
        while do_work:
            time.sleep(interval)
            if status_only:  # just show info once, don't do anything else
                do_work = False
            jq = os.listdir('{}/jobqueue'.format(self.wdir))
            jp = os.listdir('{}/jobpod'.format(self.wdir))
            wq = os.listdir('{}/workerqueue'.format(self.wdir))
            self.sort_jobqueue(jq)
            num_workers = len(os.listdir('{}/screenrcs'.format(self.wdir)))
            if len(jq) > 0:  # more jobs waiting for workers
                sys.stdout.write("\r[fjd-dispatcher] {} job(s) waiting in the queue."\
                                 " Currently {} worker(s) out of {} are free ...                   "\
                       .format(len(jq), len(wq), num_workers))
                sys.stdout.flush()
                if not status_only:
                    for _ in range(min(len(jq), len(wq))):
                        worker = wq.pop()
                        job = jq.pop()
                        os.rename('{wdir}/jobqueue/{j}'.format(wdir=self.wdir, j=job),
                                '{wdir}/jobpod/{w}'.format(wdir=self.wdir, w=worker))
                        os.remove('{wdir}/workerqueue/{w}'.format(wdir=self.wdir, w=worker))
            elif len(jp) > 0:  # some jobs are still running
                sys.stdout.write("\r[fjd-dispatcher] Job queue is empty. Waiting for remaining {} job(s) to finish ...        ".format(len(jp)))
                sys.stdout.flush()
            else:  # all jobs are done
                sys.stdout.write("\r[fjd-dispatcher] Job queue is empty and all jobs have finished.                                ")
                sys.stdout.flush()
                if end_when_jobs_are_done:
                    sys.stdout.write("\n")
                    Recruiter(project=project).fire()
                    do_work = False
                    if end_callback:
                        subprocess.call(end_callback, shell=True)
            if status_only:
                print('')

        self.wrap_up()

    def sort_jobqueue(self, jobqueue):
        '''
        Sort jobs - overwrite to enable dynamic queueing.
        Only an idea as of now, see issue #13
        '''
        pass



#!/usr/bin/python
import os
import sys
from multiprocessing import cpu_count

from fjd.core_process import CoreProcess
from fjd.utils import ensure_wdir, empty_queues
from fjd import Recruiter, Dispatcher


class Main(CoreProcess):

    def __init__(self, exe, repeat=1, parameters=[], project=None, num_workers=0, curdir=''):
        if not exe or exe == '':
            print('[fjd] Please specify an executable command (--exe).')
            sys.exit(2)
        if repeat > 1 and len(parameters) > 0:
            print('[fjd] Only one of --repeat and --parameters can be set at a time.')
            sys.exit(2)
        empty_queues(project=project)
        self.wdir = ensure_wdir(project)
        if len(parameters) > 1:
            for i, p in enumerate(parameters):
                job = '{}/jobqueue/job{}'.format(self.wdir, i)
                with open(job, 'w') as f:
                    f.write('#!/bin/bash\n')
                    cur_exe = exe
                    ext_params = []
                    for j, param in enumerate(str(p).split('#')):
                        if '${}'.format(j+1) in cur_exe:
                            cur_exe = cur_exe.replace('${}'.format(j+1), str(param))
                        else:
                            ext_params.append(str(p))
                    f.write('{exe} {params}'.format(exe=cur_exe, params=' '.join(ext_params)))
                os.chmod(job, 0777)
        else:
            for i in range(repeat):
                job = '{}/jobqueue/job{}'.format(self.wdir, i)
                with open(job, 'w') as f:
                    f.write('#!/bin/bash\n')
                    f.write(exe)
                os.chmod(job, 0777)

        if num_workers == 0:
            num_workers = cpu_count() - 1
        num_workers = min(num_workers, cpu_count())
        recruiter = Recruiter(num_workers=num_workers, project=project,
                              curdir=curdir)
        recruiter.hire()
        Dispatcher(project=project)

